#ifndef UTILS_H_
#define UTILS_H_

#include <sys/stat.h>
#include <limits.h>
#include <dirent.h>	
#include "config.h" // definition of : extern config_settings_t config used by loglevel and logv	

/***** Enumerations for use with public functions *****/

enum convert_endpoint_t {cept_client=0, cept_server, cept_config, cept_playlist};
enum list_searchmode_t {lsmt_exact=0, lsmt_begins_with, lsmt_ends_with};

// Global storage of configuration used by scandir() and scandir_selector()
typedef struct scandir_info_s {
	char   scandir_basename_pattern[PATH_MAX] ;	// ex: movie*.*
	char   scandir_file_extensions[PATH_MAX] ;	// ex: srt,apt,txt,sub...
	char   scandir_dirname[PATH_MAX] ;			// ex: /data/media/video/Children
	short  scandir_directory_limited ;			// flag 1=Only directory name is retrieved by scandir()
} scandir_info_t;
extern  scandir_info_t 	scandir_info;	// Variable set before any call to scandir()

#define SCANDIR_FILE_MODE 		0	// only files are retrieved from scandir()		
#define SCANDIR_DIRECTORY_MODE 	1	// only directory entries are retrived from scandir()


// Global storage of temporary file used by ftw() and ftw_process_file() (#ifdef USE_FTW)
typedef struct ftw_info_s {
	char   ftw_tempfilename[PATH_MAX] ;			// ex: /home/olivier/media/video/test/.piccache/tmpsEnNZH
	int    ftw_temp_fd ;						// ex: file descriptor get with fopen()
	int    ftw_nb_of_cachedfile ;			    // ex: number of files found in cached directory
} ftw_info_t;
extern ftw_info_t 	ftw_info;			// Variable set before any call to ftw()


/***** LogLevel ***/
#define LOGSTD		0
#define LOGERROR	1
#define LOGWARN		2
#define LOGPROTO	3
#define LOGINFO		4
#define LOGDEBUG	5

#define LOGNAME_STD		""
#define LOGNAME_ERROR	"Error"
#define LOGNAME_WARN	"Warning"
#define LOGNAME_PROTO	"Protocol"
#define LOGNAME_INFO	"Information"
#define LOGNAME_DEBUG	"Debug"
/***** Prototypes of public functions *****/

// Memory management

void * xmalloc(size_t size);
void * xrealloc(void *ptr, size_t size);

// Character set conversion

const char* convert_charset_direct(const char* s, const char *from_character_set, const char *to_character_set);
const char* convert_charset(const char* s, enum convert_endpoint_t source_endpoint, enum convert_endpoint_t dest_endpoint);
int can_convert_charset(enum convert_endpoint_t source_endpoint, enum convert_endpoint_t dest_endpoint);

// String handling, including list handling

void make_lowercase(char *buf);
const char* list_pointer(const char *haystack, const char *needle, const char *delimiters, enum list_searchmode_t);
int list_contains(const char *haystack, const char *needle, const char *delimiters);
int list_contains_extension(const char *list, const char *filename);
int string_substitute(char *haystack, const char *needle, const char *substitution, int haystack_size);

// Case-insensitive filename operations

int casestat(char *filename, struct stat *statbuf);
int my_filelist_compare(const void *ap, const void *bp);

// Directory browsing

int scandir_selector(const struct dirent *entry);
void setup_scandir_selector(const char *basename_pattern, const char *file_extensions, 
							const char *dirname, int directory_limited);
int ftw_extract_timestamp_file(const char *filename, const struct stat *filestat, int fileflag);
int extract_timestamp_file(const char *filename, const struct stat *filestat, int fd, const char *tempfilename);

// Networking

int receive_line(int sock, char *buffer, int size, int timeout_seconds);
int send_bytes(int sock, const char *buffer, int size);
void send_line(int sock, const char *line);

// Misc

int get_recent_path(char *recentpath, const char *mediapath);
void clean_pathname(char *pathname);
int verify_path(char *p);
int construct_pseudo_name(const char* prefix, char** thisname_p, int fullpath_len);
void rename_filetype_outgoing(char* filename);
void rename_filetype_incoming(char* filename);


#endif /*UTILS_H_*/
